__all__ = ["WFileSparseCube"]

import collections
import copy
import matplotlib.pyplot as plt
from pylab import MaxNLocator
import numbers
import numpy as np
import os
import os.path
from itertools import product, combinations, cycle
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XScaleSpectrum import *
from .a_WSparseCube import *
import a99
# from .... import explorer as ex
import f311.filetypes as ft

_COLORS_SQ = [(.1, .6, .5), (.5, .1, .7)]
_ITER_COLORS_SQ = cycle(_COLORS_SQ)


class WFileSparseCube(a99.WEditor):
    """
    FileSparseCube editor widget.

    Args:
      parent=None
    """

    def __init__(self, parent):
        from f311 import explorer as ex
        a99.WEditor.__init__(self, parent)

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        # GUI update needed but cannot be applied immediately, e.g. visible range being edited
        # each flag for one tab
        self.flag_update_pending = [False, False]
        # callables to update each visualization tab
        self.map_update_vis = [self.plot_spectra, self.plot_colors]
        # Whether there is sth in yellow background in the Headers tab
        self.flag_header_changed = False
        self.obj_square = None

        # # Central layout
        lantanide = self.layout_editor
        a99.set_margin(lantanide, 0)
        self.setLayout(lantanide)

        # ## Horizontal splitter occupying main area: (options area) | (plot area)
        sp2 = self.splitter2 = QSplitter(Qt.Horizontal)
        lantanide.addWidget(sp2)

        # ## Widget left of horizontal splitter, containing (File Line) / (Options area)
        wfilett0 = self.keep_ref(QWidget())
        lwfilett0 = QVBoxLayout(wfilett0)
        a99.set_margin(lwfilett0, 0)

        # ### Tabbed widget occupying left of horizontal splitter (OPTIONS TAB)
        tt0 = self.tabWidgetOptions = QTabWidget(self)
        lwfilett0.addWidget(tt0)
        tt0.setFont(a99.MONO_FONT)
        tt0.currentChanged.connect(self.current_tab_changed_options)

        # #### Tab: Vertical Splitter between "Place Spectrum" and "Existing Spectra"
        spp = QSplitter(Qt.Vertical)
        tt0.addTab(spp, "&Spectra")

        # ##### Existing Spectra area
        wex = QWidget()
        lwex = QVBoxLayout(wex)
        a99.set_margin(lwex, 3)
        ###
        w = self.wsptable = WSparseCube(self.parent_form)
        w.changed.connect(self.on_spectra_edited)
        lwex.addWidget(w)

        # ##### Finally...
        spp.addWidget(wex)

        # #### Second tab (NEW FileSparseCube)
        sa1 = self.keep_ref(QScrollArea())
        tt0.addTab(sa1, "&Header")
        sa1.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        sa1.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Widget that will be handled by the scrollable area
        w = self.keep_ref(QWidget())
        sa1.setWidget(w)
        sa1.setWidgetResizable(True)
        lscrw = QVBoxLayout(w)
        a99.set_margin(lscrw, 3)
        ###
        lscrw.addWidget(self.keep_ref(QLabel("<b>Header properties</b>")))
        ###
        b = self.keep_ref(QPushButton("Collect field names"))
        b.clicked.connect(self.on_collect_fieldnames)
        lscrw.addWidget(b)

        # Form layout
        lg = self.keep_ref(QGridLayout())
        a99.set_margin(lg, 0)
        lg.setVerticalSpacing(4)
        lg.setHorizontalSpacing(5)
        lscrw.addLayout(lg)
        ###
        lscrw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        # field map: [(label widget, edit widget, field name, short description, long description), ...]
        map = self._map1 = []

        ###
        x = self.keep_ref(QLabel())
        y = self.edit_fieldnames = QPlainTextEdit()
        y.textChanged.connect(self.on_header_edited)
        x.setBuddy(y)
        map.append((x, y, "&Field names", "'header' information for each spectrum", "", lambda: self.f.sparsecube.fieldnames,
                   lambda: self.edit_fieldnames.toPlainText()))
        ###
        x = self.label_width = QLabel()
        y = self.spinbox_width = QSpinBox()
        y.valueChanged.connect(self.on_header_edited)
        y.setMinimum(1)
        y.setMaximum(1000)
        x.setBuddy(y)
        map.append((x, y, "&width", "hi-resolution (HR) width (pixels)", "", lambda: self.f.sparsecube.width,
                   lambda: self.spinbox_width.value()))
        ###
        x = self.label_height = QLabel()
        y = self.spinbox_height = QSpinBox()
        y.valueChanged.connect(self.on_header_edited)
        y.setMinimum(0)
        y.setMaximum(1000)
        x.setBuddy(y)
        map.append(
            (x, y, "&height", "HR height (pixels)", "", lambda: self.f.sparsecube.height, lambda: self.spinbox_height.value()))
        ###
        x = self.label_hrfactor = QLabel()
        y = self.spinbox_hrfactor = QSpinBox()
        y.valueChanged.connect(self.on_header_edited)
        y.setMinimum(1)
        y.setMaximum(100)
        x.setBuddy(y)
        map.append((x, y, "&hrfactor", "(HR width)/(ifu width)", "", lambda: self.f.sparsecube.hrfactor,
                   lambda: self.spinbox_hrfactor.value()))
        ###
        x = self.label_hr_pix_size = QLabel()
        y = self.lineEdit_hr_pix_size = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_header_edited)
        y.setValidator(QDoubleValidator(0, 1, 7))
        x.setBuddy(y)
        map.append((x, y, "&hr_pix_size", "HR pixel width/height (arcsec)", "", lambda: self.f.sparsecube.hr_pix_size,
                   lambda: float(self.lineEdit_hr_pix_size.text())))
        ###
        x = self.label_hrfactor = QLabel()
        y = self.spinbox_R = QSpinBox()
        y.valueChanged.connect(self.on_header_edited)
        y.setMinimum(100)
        y.setMaximum(100000)
        y.setSingleStep(100)
        x.setBuddy(y)
        map.append(
            (x, y, "&R", "resolution (delta lambda)/lambda", "", lambda: self.f.sparsecube.R, lambda: self.spinbox_R.value()))

        for i, (label, edit, name, short_descr, long_descr, f_from_f, f_from_edit) in enumerate(map):
            # label.setStyleSheet("QLabel {text-align: right}")
            assert isinstance(label, QLabel)
            label.setText(a99.enc_name_descr(name, short_descr))
            label.setAlignment(Qt.AlignRight)
            lg.addWidget(label, i, 0)
            lg.addWidget(edit, i, 1)
            label.setToolTip(long_descr)
            edit.setToolTip(long_descr)

        lgo = QHBoxLayout()
        a99.set_margin(lgo, 0)
        lscrw.addLayout(lgo)
        ###
        bgo = self.button_revert = QPushButton("Revert")
        lgo.addWidget(bgo)
        bgo.clicked.connect(self.header_revert)
        ###
        bgo = self.button_apply = QPushButton("Apply")
        lgo.addWidget(bgo)
        bgo.clicked.connect(self.header_apply)
        ###
        lgo.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # TODO get rid of this
        # #### More Tools tab
        wset = self.keep_ref(QWidget())
        tt0.addTab(wset, "&More")
        lwset = self.keep_ref(QVBoxLayout(wset))
        ###
        b = self.keep_ref(QPushButton("&Crop in new window..."))
        lwset.addWidget(b)
        b.clicked.connect(self.crop_clicked)
        ###
        b = self.keep_ref(QPushButton("E&xport %s ..." % ft.FileFullCube.description))
        lwset.addWidget(b)
        b.clicked.connect(self.export_ccube_clicked)
        ###
        lwset.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        # ### Tabbed widget occupying right of horizontal splitter
        tt1 = self.tabWidgetVis = QTabWidget(self)
        tt1.setFont(a99.MONO_FONT)
        tt1.currentChanged.connect(self.current_tab_changed_vis)

        # #### Tab containing 3D plot representation
        w0 = self.keep_ref(QWidget())
        tt1.addTab(w0, "&Plot 3D")
        # http://stackoverflow.com/questions/12459811
        self.figure0, self.canvas0, self.lfig0 = a99.get_matplotlib_layout(w0)

        # lscrw.addLayout(lfig)

        # #### Colors tab
        w1 = self.keep_ref(QWidget())
        tt1.addTab(w1, "&Colors")
        ###
        lw1 = QVBoxLayout(w1)
        lwset = self.keep_ref(QHBoxLayout())
        lw1.addLayout(lwset)
        ###
        la = self.keep_ref(QLabel("&Visible range"))
        lwset.addWidget(la)
        ###
        ed = self.lineEdit_visibleRange = QLineEdit("[3800., 7500.]")
        lwset.addWidget(ed)
        la.setBuddy(ed)
        ed.textEdited.connect(self.on_colors_setup_edited)
        ###
        la = self.keep_ref(QLabel("Color map"))
        lwset.addWidget(la)
        ###
        ed = self.comboBox_cmap = QComboBox()
        ed.addItem("0-Rainbow")
        ed.addItem("1-RGB")
        ed.setCurrentIndex(0)
        lwset.addWidget(ed)
        la.setBuddy(ed)
        ed.currentIndexChanged.connect(self.on_colors_setup_edited)
        ###
        cb = self.checkBox_scale = QCheckBox("Scale colors")
        lwset.addWidget(cb)
        # cb.setTooltip("If checked, will make color luminosity proportional to flux area under the visible range")
        cb.stateChanged.connect(self.on_colors_setup_edited)
        ###
        b = self.keep_ref(QPushButton("Redra&w"))
        lwset.addWidget(b)
        b.clicked.connect(self.replot_colors)
        ###
        lwset.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        ###
        wm = self.keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        lw1.addWidget(wm)
        self.figure1, self.canvas1, self.lfig1 = a99.get_matplotlib_layout(wm)
        self.canvas1.mpl_connect('button_press_event', self.on_colors_click)

        # ### Finally ...
        sp2.addWidget(wfilett0)
        sp2.addWidget(tt1)

        # # Timers
        ##########
        t = self.timer_place = QTimer(self)
        t.timeout.connect(self.timeout_place)
        t.setInterval(500)
        t.start()

        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True
        self.add_log("Welcome from %s.__init__()" % (self.__class__.__name__))

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def _do_load(self, x):
        assert isinstance(x, (ft.FileSparseCube, ft.FileFullCube))

        # Converts from FileFullCube to FileSparseCube format, if necessary"""
        x1 = None
        if isinstance(x, ft.FileFullCube):
            x1 = ft.FileSparseCube()
            x1.sparsecube.from_full_cube(x.wcube)
        if x1:
            x1.filename = a99.add_bits_to_path(x.filename, "imported-from-",
             os.path.splitext(ft.FileSparseCube.default_filename)[1])
            x = x1

        self._f = x
        self.wsptable.set_collection(x.sparsecube)
        self.__update_gui(True)
        self._flag_valid = True  # assuming that file does not come with errors
        self.setEnabled(True)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        # TODO self.lineEdit_titrav.setFocus()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_colors_setup_edited(self):
        if self.flag_process_changes:
            self.flag_update_pending[1] = True

    def on_header_edited(self):
        if self.flag_process_changes:
            sth = False
            sndr = self.sender()
            for _, edit, _, _, _, f_from_f, f_from_edit in self._map1:
                changed = f_from_f() != f_from_edit()
                sth = sth or changed
                if edit == sndr:
                    a99.style_widget_changed(self.sender(), changed)
            self.set_flag_header_changed(sth)

    def on_spectra_edited(self):
        try:
            self.__update_gui_vis()
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True, e)
        self.changed.emit()

    def on_place_spectrum_edited(self):
        # could only update the obj_square but this is easier
        self.plot_colors()

    def header_revert(self):
        self.__update_gui_header()

    def header_apply(self):
        if self.__update_f_header(self._f.sparsecube):
            self.__update_gui(True)

    def current_tab_changed_vis(self):
        self.__update_gui_vis_if_pending()

    def current_tab_changed_options(self):
        pass

    def timeout_place(self):
        if self.obj_square:
            next_color = next(_ITER_COLORS_SQ)
            for obj in self.obj_square:
                obj.set_color(next_color)
                self.canvas1.draw()

    # TODO don't worry, this shouldn't be here anyway
    def crop_clicked(self):
        try:
            sky = self._f.sparsecube

            specs = (("x_range", {"value": "[%d, %d]" % (0, sky.width - 1)}),
                     ("y_range", {"value": "[%d, %d]" % (0, sky.height - 1)}),
                     ("wavelength_range", {"value": "[%g, %g]" % (sky.wavelength[0], sky.wavelength[-1])})
                     )
            # fields = ["x_range", "y_range", "wavelength_range"]
            form = a99.XParametersEditor(specs=specs, title="Select sub-cube")
            while True:
                r = form.exec_()
                if not r:
                    break
                kk = form.get_kwargs()
                s = ""
                try:
                    s = "x_range"
                    x0, x1 = eval(kk["x_range"])
                    s = "y_range"
                    y0, y1 = eval(kk["y_range"])
                    s = "wavelength_range"
                    lambda0, lambda1 = eval(kk["wavelength_range"])
                except Exception as E:
                    self.add_log_error("Failed evaluating %s: %s" % (s, a99.str_exc(E)), True)
                    continue

                # Works with clone, then replaces original, to ensure atomic operation
                clone = copy.deepcopy(self._f)
                clone.filename = None
                try:
                    clone.sparsecube.crop(x0, x1, y0, y1, lambda0, lambda1)
                except Exception as E:
                    self.add_log_error("Crop operation failed: %s" % a99.str_exc(E), True)
                    continue

                form1 = self.keep_ref(self.parent_form.__class__())
                form1.load(clone)
                form1.show()

                # # Replaces original
                # self._f = clone
                # self.__update_gui(True)
                break

        except Exception as E:
            self.add_log_error("Crop failed: %s" % a99.str_exc(E), True, E)

    def export_ccube_clicked(self):
        fn = QFileDialog.getSaveFileName(self, "Save file in %s format" % ft.FileFullCube.description,
                                         ft.FileFullCube.default_filename, "*.fits")[0]
        if fn:
            try:
                fn = str(fn)
                wcube = self._f.sparsecube.to_full_cube()
                fccube = ft.FileFullCube()
                fccube.wcube = wcube
                fccube.save_as(fn)
            except Exception as E:
                self.add_log_error("Failed export: %s" % a99.str_exc(E), True, E)

    def replot_colors(self):
        self.plot_colors()

    def on_colors_click(self, event):
        x, y = int(event.xdata + .5), int(event.ydata + .5)
        if 0 <= x < self._f.sparsecube.width and 0 <= y < self._f.sparsecube.height:
            self.wsptable.spinbox_x.setValue(x)
            self.wsptable.spinbox_y.setValue(y)
            self.plot_colors()

    def on_collect_fieldnames(self):
        # TODO confirmation
        self.edit_fieldnames.setPlainText(str(self._f.sparsecube.collect_fieldnames()))

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __emit_if(self):
        if self.flag_process_changes:
            self.changed.emit()

    def __update_gui(self, flag_header=False):
        """Updates GUI to reflect what is in self._f"""
        self.flag_process_changes = False
        try:
            self.wsptable.update()
            self.__update_gui_vis()
            if flag_header:
                self.__update_gui_header()
        finally:
            self.flag_process_changes = True

    def __update_gui_vis(self):
        idx = self.tabWidgetVis.currentIndex()
        for i, callable_ in enumerate(self.map_update_vis):
            # Updates current visualization tab and flags update pending for other tabs
            if i == idx:
                callable_()
                self.flag_update_pending[i] = False
            else:
                self.flag_update_pending[i] = True

    def __update_gui_header(self):
        """Updates header controls only"""
        sky = self._f.sparsecube
        self.spinbox_width.setValue(sky.width)
        self.spinbox_height.setValue(sky.height)
        self.spinbox_hrfactor.setValue(sky.hrfactor)
        self.lineEdit_hr_pix_size.setText(str(sky.hr_pix_size))
        self.spinbox_R.setValue(sky.R)
        self.edit_fieldnames.setPlainText(str(sky.fieldnames))
        self.set_flag_header_changed(False)

    def set_flag_header_changed(self, flag):
        self.button_apply.setEnabled(flag)
        self.button_revert.setEnabled(flag)
        self.flag_header_changed = flag
        if not flag:
            # If not changed, removes all eventual yellows
            for _, edit, _, _, _, _, _ in self._map1:
                a99.style_widget_changed(edit, False)

    def _update_fobj(self):
        sky = self._f.sparsecube
        self._flag_valid = self.__update_f_header(sky)

    def __update_f_header(self, sky):
        """Updates headers of a SparseCube objects using contents of the Headers tab"""
        emsg, flag_error = "", False
        ss = ""
        try:
            ss = "fieldnames"
            ff = a99.eval_fieldnames(str(self.edit_fieldnames.toPlainText()))
            sky.fieldnames = ff
            ss = "width"
            sky.width = int(self.spinbox_width.value())
            ss = "height"
            sky.height = int(self.spinbox_height.value())
            ss = "hrfactor"
            sky.hrfactor = int(self.spinbox_hrfactor.value())
            ss = "hr_pix_size"
            sky.hr_pix_size = float(self.lineEdit_hr_pix_size.text())
            ss = "R"
            sky.R = float(self.spinbox_R.value())
            self.__update_gui(True)
            flag_emit = True
        except Exception as E:
            flag_error = True
            if ss:
                emsg = "Field '%s': %s" % (ss, a99.str_exc(E))
            else:
                emsg = a99.str_exc(E)
            self.add_log_error(emsg)
        if flag_emit:
            self.__emit_if()
        return not flag_error

    def __update_gui_vis_if_pending(self):
        idx = self.tabWidgetVis.currentIndex()
        if self.flag_update_pending[idx]:
            self.map_update_vis[idx]()
            self.flag_update_pending[idx] = False

    def plot_spectra(self):
        from f311 import explorer as ex
        # self.clear_markers()
        if self._f is None:
            return

        try:
            fig = self.figure0
            fig.clear()
            ax = fig.gca(projection='3d')
            ex.draw_cube_3d(ax, self._f.sparsecube)
            fig.tight_layout()
            self.canvas0.draw()

        except Exception as E:
            self.add_log_error(a99.str_exc(E))
            a99.get_python_logger().exception("Could not plot spectra")

    def plot_colors(self):
        from f311 import explorer as ex
        # self.clear_markers()
        if self._f is None:
            return

        try:
            vrange = eval(str(self.lineEdit_visibleRange.text()))
            if len(vrange) != 2 or not all([isinstance(x, numbers.Number) for x in vrange]):
                raise RuntimeError('Visible range must be a sequence with two numbers')

            fig = self.figure1
            fig.clear()
            ax = fig.gca()
            sqx, sqy = None, None
            flag_scale = self.checkBox_scale.isChecked()
            method = self.comboBox_cmap.currentIndex()
            try:
                sqx, sqy = self.wsptable.get_pixel_xy()
            except:
                a99.get_python_logger().exception("Cannot draw square")
                pass  # Nevermind (does not draw square)
            self.obj_square = ex.draw_cube_colors(ax, self._f.sparsecube, vrange, sqx, sqy, flag_scale, method)

            fig.tight_layout()
            self.canvas1.draw()

            self.flag_plot_colors_pending = False
        except Exception as E:
            self.add_log_error(a99.str_exc(E))
            a99.get_python_logger().exception("Could not plot colors")




__all__ = ["XFileSpectrumList"]


from .a_WFileSpectrumList import *
import a99
import f311.filetypes as ft
from ..a_XFileMainWindow import *


class XFileSpectrumList(XFileMainWindow):
    def _add_stuff(self):
        import f311

        self.setWindowTitle(a99.get_window_title("Spectrum List Editor"))

        ce = self.ce = WFileSpectrumList(self)

        self.pages.append(MyPage(text_tab="FileSpectrumList editor", cls_save=ft.FileSpectrumList,
            clss_load=[ft.FileSpectrumList, ft.FileFullCube]+f311.classes_sp(), wild="*.splist",
            editor=ce))

        # # Adds spectrum collection actions to menu
        self.menuBar().addMenu(self.ce.menu_actions)



"""Plots ESO sky model"""
from astropy.io import fits
import matplotlib.pyplot as plt


def load_eso_sky(filename):
    """Loads ESO sky model and returns (lambda, emission, transmission)"""

    hl = fits.open("eso-sky.fits")
    d = hl[1].data

    return d["lam"]*10000, d["flux"], d["trans"]


# lam:     vacuum wavelength in micron
# flux:    sky emission radiance flux in ph/s/m2/micron/arcsec2
# dflux1:  sky emission -1sigma flux uncertainty
# dflux2:  sky emission +1sigma flux uncertainty
# dtrans:  sky transmission
# dtrans1: sky transmission -1sigma uncertainty
# dtrans2: sky transmission +1sigma uncertainty


hl = fits.open("eso-sky.fits")
d = hl[1].data

x, em, tr = load_eso_sky("eso-sky.fits")
ax = plt.subplot(211)
plt.plot(x, em, "k")
plt.title("Emission")

plt.subplot(212, sharex=ax)
plt.plot(x, tr, "k")
plt.title("Transmission")
plt.xlabel("Wavelength ($\AA$)")
plt.xlim(x[[0, -1]])


plt.tight_layout()
plt.show()




# # Version 1
# # dflux1, dflux2, dtrans1, dtrans2 didn't seem to be that useful
# hl = fits.open("eso-sky.fits")
# d = hl[1].data
#
# x = d["lam"]
#
# plt.subplot(211)
# plt.plot(x, d["flux"], "k", lw=2)
# plt.plot(x, d["dflux1"], "k--")
# plt.plot(x, d["dflux2"], "k--")
#
# plt.subplot(212)
# plt.plot(x, d["trans"], "k", lw=2)
# plt.plot(x, d["dtrans1"], "k--")
# plt.plot(x, d["dtrans2"], "k--")
#
# plt.show()

# # Version 0
# names = [column.name for column in d.columns]
# nc = len(names)
#
# x = d[names[0]]
#
# for name in names[1:]:
#     y = d[name]
#     plt.plot(x, y)
#
# plt.show()


__all__ = ["WSparseCube"]

import copy
import os
import os.path
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XScaleSpectrum import *
from astropy import units as u
import a99
# from .... import explorer as ex
import f311.filetypes as ft
from .a_WSpectrumCollectionBase import *
import f311


class WSparseCube(WSpectrumCollectionBase):
    """Editor for SparseCube objects"""

    # argument0: flag_changed_header
    changed = pyqtSignal(bool)

    def __init__(self, parent):
        WSpectrumCollectionBase.__init__(self, parent, True)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False

        # # Creates actions

        # ## Makes a menu
        menu = self.menu_actions = QMenu("Spectrum &List")
        menu.addAction(self.action_add_spectra)
        m = keep_ref(menu.addMenu("With &All Spectra"))
        # m.addAction(self.action_sel_use_spectrum_block)
        m.addAction(self.action_all_to_scalar)
        m.addAction(self.action_all_plot_xy)
        m.addAction(self.action_all_plot_xyz)
        m.addAction(self.action_all_export_csv)
        m = keep_ref(menu.addMenu("With &Selected Spectra"))
        m.addAction(self.action_sel_use_spectrum_block)
        m.addAction(self.action_sel_plot_stacked)
        m.addAction(self.action_sel_plot_overlapped)
        m.addAction(self.action_sel_open_in_new)
        m.addAction(self.action_sel_delete)
        m = keep_ref(menu.addMenu("With &Current Spectrum"))
        m.addAction(self.action_curr_scale)


        # # Central layout
        # Will have a toolbox and a table. When the toolbox contracts, there is more space for the table
        lwmain = self.layout_editor
        a99.set_margin(lwmain, 0)
        self.setLayout(lwmain)

        lxy = self.keep_ref(QHBoxLayout())
        lwmain.addLayout(lxy)


        x = self.keep_ref(QLabel("<b>To add spectra:</b>"))
        lxy.addWidget(x)

        x = self.label_x = QLabel("x-coordinate")
        y = self.spinbox_x = QSpinBox()
        y.setToolTip("x-coordinate to add spectra to cube (in pixels; 0-based)")
        # y.valueChanged.connect(self.on_place_spectrum_edited)
        y.setMinimum(0)
        y.setMaximum(1000)
        x.setBuddy(y)
        lxy.addWidget(x)
        lxy.addWidget(y)

        x = self.label_y = QLabel("x-coordinate")
        y = self.spinbox_y = QSpinBox()
        y.setToolTip("y-coordinate to add spectra to cube (in pixels; 0-based)")
        # y.valueChanged.connect(self.on_place_spectrum_edited)
        y.setMinimum(0)
        y.setMaximum(1000)
        x.setBuddy(y)
        lxy.addWidget(x)
        lxy.addWidget(y)

        lxy.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ## Toolbox
        # The toolbox will have only one widget whose layout is a grid with two columns:
        #   - The first column contains labels
        #   - The second column contains panels of buttons layed horizontally


        wtoolboxes = keep_ref(a99.WCollapsiblePanel())
        lwmain.addWidget(wtoolboxes)
        wtoolboxes.label.setText("<b>Tools</b>")
        wtoolboxes.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # ### Layout to have (grid layout, spacer)
        lwtoolboxes = QHBoxLayout(wtoolboxes.widget)
        lwmain.addLayout(lwtoolboxes)
        a99.set_margin(lwtoolboxes, 0)
        lwtoolboxes.setSpacing(0)

        # #### Grid layour
        lg = QGridLayout()
        lwtoolboxes.addLayout(lg)
        a99.set_margin(lg, 2)
        lg.setSpacing(0)
        label = keep_ref(QLabel('<b>File:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 0, 0)
        label = keep_ref(QLabel('<b>With all:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 1, 0)
        label = keep_ref(QLabel('<b>With selected:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 2, 0)
        label = keep_ref(QLabel('<b>With current:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 3, 0)


        TOOL_BUTTON_STYLE = Qt.ToolButtonTextBesideIcon
        # TOOL_BUTTON_STYLE = Qt.ToolButtonIconOnly


        # ##### Toolbar #0: operations independent of the table widget selection state
        ###
        tb = self.toolBar0 = QToolBar()
        lg.addWidget(tb, 0, 1)
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_add_spectra)
        tb.setIconSize(QSize(16, 16))

        # ##### Toolbar #1: operations independent of the table widget selection state
        ###
        tb = self.toolBar1 = QToolBar()
        lg.addWidget(tb, 1, 1)
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.setIconSize(QSize(16, 16))
        tb.addAction(self.action_all_to_scalar)
        tb.addAction(self.action_all_plot_xy)
        tb.addAction(self.action_all_plot_xyz)
        tb.addAction(self.action_all_export_csv)


        # ##### Toolbar #2: operations affecting only the spectra which are selected
        ###
        tb = self.toolBar2 = QToolBar()
        lg.addWidget(tb, 2, 1)
        tb.setIconSize(QSize(16, 16))
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_sel_use_spectrum_block)
        tb.addAction(self.action_sel_plot_stacked)
        tb.addAction(self.action_sel_plot_overlapped)
        tb.addAction(self.action_sel_open_in_new)
        tb.addAction(self.action_sel_delete)

        # ##### Toolbar #3: operations affecting only the current spectrum
        ###

        tb = self.toolBar3 = QToolBar()
        lg.addWidget(tb, 3, 1)
        tb.setIconSize(QSize(16, 16))
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_curr_scale)

        # #### Spacer
        lwtoolboxes.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # ## Table Widget
        a = self.twSpectra = QTableWidget()
        lwmain.addWidget(a)
        a.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        a.setSelectionMode(QAbstractItemView.MultiSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setAlternatingRowColors(True)
        a.setEditTriggers(QAbstractItemView.DoubleClicked | QAbstractItemView.EditKeyPressed)
        a.setFont(a99.MONO_FONT)
        a.installEventFilter(self)
        a.setContextMenuPolicy(Qt.CustomContextMenu)
        a.customContextMenuRequested.connect(self.on_twSpectra_customContextMenuRequested)
        a.setSortingEnabled(True)
        a.cellChanged.connect(self.on_twSpectra_cellChanged)
        a.itemSelectionChanged.connect(self.on_twSpectra_itemSelectionChanged)
        ah = a.horizontalHeader()
        ah.setSectionsMovable(True)
        ah.setContextMenuPolicy(Qt.CustomContextMenu)
        ah.customContextMenuRequested.connect(self.show_header_context_menu)
        ah.setSelectionMode(QAbstractItemView.SingleSelection)
        ah.sectionMoved.connect(self.section_moved)


        # # Final adjustments
        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True
        self.add_log("Welcome from %s.__init__()" % (self.__class__.__name__))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def get_pixel_xy(self):
        x = int(self.spinbox_x.value())
        if not (0 <= x < self.collection.width):
            raise RuntimeError("x must be in [0, %s)" % self.collection.width)
        y = int(self.spinbox_y.value())
        if not (0 <= y < self.collection.height):
            raise RuntimeError("y must be in [0, %s)" % self.collection.height)
        return x, y


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_add_spectra(self):
        flag_emit = False
        filenames = QFileDialog.getOpenFileNames(self, "Add Spectra", "",
         "All files(*.*);;Spectrum List files (*.splist);;Sparse Cube files (*.sparsecube)")[0]
        if not filenames:
            return

        # classes = f311.classes_sp()+[ft.FileSpectrumList, ft.FileSparseCube]
        classes = f311.classes_collection()
        report, successful, failed = ["<b>Results</b>"], [], []
        for filename in filenames:
            filename = str(filename)
            basename = os.path.basename(filename)
            file = ft.load_with_classes(filename, classes)
            try:
                if file is None:
                    raise RuntimeError("Could not load file")
                if isinstance(file, ft.FileSpectrum):
                    self.collection.add_spectrum(file.spectrum, *self.get_pixel_xy())
                elif isinstance(file, ft.FileSpectrumList):
                    self.collection.merge_with(file.splist, *self.get_pixel_xy())
                elif isinstance(file, ft.FileSparseCube):
                    self.collection.merge_with(file.sparsecube)
                successful.append("  - %s" % basename)
            except Exception as e:
                failed.append('&nbsp;&nbsp;- %s: %s' % (basename, str(e)))
                s = "Error adding file '%s': %s" % (basename, a99.str_exc(e))
                a99.get_python_logger().exception(s)
                self.add_log_error(s)

        if len(successful) > 0:
            report.extend(["", "Successful:"])
            report.extend(successful)

            # # "Gambiarra" to expose field names
            # if not self.is_spectrum_list():
            #     c = self.collection
            #     ffnn, aann = ["PIXEL-X", "PIXEL-Y"], ["fieldnames", "fieldnames_visible"]
            #     for fn in ffnn:
            #         for an in aann:
            #             a = c.__getattribute__(an)
            #             if fn not in a:
            #                 a.append(fn)

            self._update_gui()
            flag_emit = True

        if len(failed) > 0:
            report.extend(["", "Failed:"])
            report.extend(failed)

        if flag_emit:
            self.changed.emit(False)

        a99.show_message("<br>".join(report))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear


    def _update_enabled_actions(self):
        has_any = self.twSpectra.rowCount() > 0
        any_selected = len(self.twSpectra.selectedItems()) > 0
        has_current = self.twSpectra.currentRow() > -1

        self.action_add_spectra.setEnabled(True)
        self.action_all_to_scalar.setEnabled(has_any)
        self.action_all_export_csv.setEnabled(has_any)
        self.action_sel_use_spectrum_block.setEnabled(any_selected)
        self.action_sel_plot_stacked.setEnabled(any_selected)
        self.action_sel_plot_overlapped.setEnabled(any_selected)
        self.action_sel_open_in_new.setEnabled(any_selected)
        self.action_sel_delete.setEnabled(any_selected)
        self.action_curr_scale.setEnabled(has_current)



__all__ = ["XPlotXYZ"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import a99
import f311.filetypes as ft


class XPlotXYZ(a99.XLogMainWindow):
    """
    Plots two fields of a SpectrumCollection object in a simple x-y plot

    Args:
      collection: SpectrumCollection object
    """
    def __init__(self, collection, *args):
        a99.XLogMainWindow.__init__(self, *args)

        self._refs = []
        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        assert isinstance(collection, ft.SpectrumCollection)
        self.collection = collection

        lw1 = keep_ref(QVBoxLayout())

        lwset = keep_ref(QHBoxLayout())
        lw1.addLayout(lwset)
        ###
        laa = keep_ref(QLabel("&X-axis"))
        lwset.addWidget(laa)
        ###
        cbx = self.comboBoxX = QComboBox()
        cbx.addItems(collection.fieldnames)
        laa.setBuddy(cbx)
        lwset.addWidget(cbx)
        ###
        laa = keep_ref(QLabel("&Y-axis"))
        lwset.addWidget(laa)
        ###
        cby = self.comboBoxY = QComboBox()
        cby.addItems(collection.fieldnames)
        laa.setBuddy(cby)
        lwset.addWidget(cby)
        ###
        laa = keep_ref(QLabel("&Z-axis"))
        lwset.addWidget(laa)
        ###
        cbz = self.comboBoxZ = QComboBox()
        cbz.addItems(collection.fieldnames)
        laa.setBuddy(cbz)
        lwset.addWidget(cbz)
        ###
        b = keep_ref(QPushButton("Re&draw"))
        lwset.addWidget(b)
        b.clicked.connect(self.redraw)
        ###
        lwset.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        ###
        wm = keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        lw1.addWidget(wm)
        self.figure, self.canvas, self.lfig = a99.get_matplotlib_layout(wm)

        cw = self.centralWidget = QWidget()
        cw.setLayout(lw1)
        self.setCentralWidget(cw)
        self.setWindowTitle("Plot Two Spectrum Collection Fields as a X-Y Chart")

        # self.redraw()

    # def get_redshift(self):
    #     return float(self.spinBox_redshift.value())


    # def spinBoxValueChanged(self, *args):
    #     self.label_redshiftValue.setText(str(self.get_redshift()))


    def redraw(self):
        try:
            fig = self.figure
            fig.clear()
            ax = fig.gca(projection='3d')

            fieldname_x = str(self.comboBoxX.currentText())
            fieldname_y = str(self.comboBoxY.currentText())
            fieldname_z = str(self.comboBoxZ.currentText())

            spectra = self.collection.spectra
            n = len(spectra)
            xx, yy, zz = np.zeros((n,)), np.zeros((n,)), np.zeros((n,))
            for i, sp in enumerate(spectra):
                s_x = sp.more_headers.get(fieldname_x)
                s_y = sp.more_headers.get(fieldname_y)
                s_z = sp.more_headers.get(fieldname_z)

                # Will ignore error converting to float
                # Cells giving error will be zero
                try:
                    xx[i] = float(s_x)
                except TypeError:
                    pass
                try:
                    yy[i] = float(s_y)
                except TypeError:
                    pass
                try:
                    zz[i] = float(s_z)
                except TypeError:
                    pass

            # 3D plot will be a scatter plot, no need to sort; besides, should we sort using X or Y?
            #
            # sort_idxs = np.argsort(xx)
            # xx = xx[sort_idxs]
            # yy = yy[sort_idxs]
            # zz = zz[sort_idxs]

            ax.plot3D(xx, yy, zz, 'ok', lw=2)
            plt.xlabel(fieldname_x)
            plt.ylabel(fieldname_y)
            plt.ylabel(fieldname_z)
            a99.format_BLB()
            self.canvas.draw()

        except Exception as e:
            self.add_log_error("Could draw figure: "+a99.str_exc(e), True)



import re
import a99

__all__ = ["compile_simids"]

def compile_simids(specs):
    """
    Compiles a list of simulation IDs (e.g., 'C000793') from a sequence of specifications

    Args:
        specs: ``[spec0, spec1, ...]``, where each element may be:

            - a number (convertible to int), or

            - a range such as '1000-1010' (string)

    Returns: list of strings starting with "C"
    """
    numbers = []
    for candidate in specs:
        try:
            if '-' in candidate:
                groups = re.match('(\d+)\s*-\s*(\d+)$', candidate)
                if groups is None:
                    raise RuntimeError("Could not parse range")
                n0 = int(groups.groups()[0])
                n1 = int(groups.groups()[1])
                numbers.extend(list(range(n0, n1 + 1)))
            else:
                numbers.append(int(candidate))
        except Exception as E:
            a99.get_python_logger().info("SKIPPED Argument '%s': %s" % (candidate, str(E)))
    numbers = set(numbers)
    simids = ["C%06d" % n for n in numbers]

    return simids



"""
Miscellanea routines that depend on other modules and sub-packages.

Rule: only 'gui/' modules can import util!!!
"""

import numpy as np
import copy
from .. import physics as ph
import a99

__all__ = ["cut_spectrum", "crop_splist",]
def cut_spectrum(sp, l0, lf):
    """
    Cuts spectrum given a wavelength interval, leaving origina intact

    Args:
        sp: Spectrum instance
        l0: initial wavelength
        lf: final wavelength

    Returns:
        Spectrum: cut spectrum
    """

    if l0 >= lf:
        raise ValueError("l0 must be lower than lf")
    idx0 = np.argmin(np.abs(sp.x - l0))
    idx1 = np.argmin(np.abs(sp.x - lf))
    out = copy.deepcopy(sp)
    out.x = out.x[idx0:idx1]
    out.y = out.y[idx0:idx1]
    return out


def get_rgb(self, visible_range=None, method=0):
    """Takes weighted average of rainbow colors RGB's

    Args:
        visible_range=None: if passed, affine-transforms the rainbow colors
        method --
          0: rainbow colors
          1: RGB
    """

    if len(visible_range) < 2:
        raise RuntimeError("Invalid visible range: {0!s}".format(visible_range))
    if visible_range[1] <= visible_range[0]:
        raise RuntimeError(
            "Second element of visible range ({0!s}) must be greater than first element".format(
                visible_range))

    if method == 1:
        # new color system splitting visible range in three
        dl = float(visible_range[1] - visible_range[0]) / 3
        ranges = np.array(
            [[visible_range[0] + i * dl, visible_range[0] + (i + 1) * dl] for i in range(3)])
        colors = np.array([(0., 0., 1.), (0., 1., 0.), (1., 0., 0.)])  # blue, green, red

        tot_area, tot_sum = 0., np.zeros(3)
        for color, range_ in zip(colors, ranges):
            b = np.logical_and(self.x >= range_[0], self.x <= range_[1])
            area = np.sum(self.y[b])
            tot_area += area
            tot_sum += color * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret

    elif method == 0:
        tot_area, tot_sum = 0., np.zeros(3)

        def ftrans(x):
            return x

        if visible_range:
            ll0 = ph.rainbow_colors[0].l0
            llf = ph.rainbow_colors[-1].lf
            ftrans = lambda lold: visible_range[0] + (visible_range[1] - visible_range[0]) / (
            llf - ll0) * (lold - ll0)

        for color in ph.rainbow_colors:
            b = np.logical_and(self.x >= ftrans(color.l0), self.x <= ftrans(color.lf))
            area = np.sum(self.y[b])
            tot_area += area
            tot_sum += color.rgb * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret
    else:
        raise RuntimeError("Unknown method: {0!s}".format(method))


def crop_splist(splist, lambda0=None, lambda1=None):
    """
    Cuts all spectra in SpectrumList

    **Note** lambda1 **included** in interval (not pythonic).
    """
    if len(splist.spectra) == 0:
        raise RuntimeError("Need at least one spectrum added in order to crop")

    if lambda0 is None:
        lambda0 = splist.wavelength[0]
    if lambda1 is None:
        lambda1 = splist.wavelength[-1]
    if not (lambda0 <= lambda1):
        raise RuntimeError('lambda0 must be <= lambda1')

    if not any([lambda0 != splist.wavelength[0], lambda1 != splist.wavelength[-1]]):
        return

    for i in range(len(splist)):
        sp = cut_spectrum(splist.spectra[i], lambda0, lambda1)
        if i == 0:
            n = len(sp)
            if n < 2:
                raise RuntimeError(
                    "Cannot cut, spectrum will have %d point%s" % (n, "" if n == 1 else "s"))
        splist.spectra[i] = sp

    splist.__update()



__all__ = ["symbols", "SYMBOLS", "C", "H"]


# All values in CGS
class LightSpeed(float):
    """Light speed in cm/s (CGS) units"""

class Planck(float):
    """Planck's constant in cm**2*g/s"""

#: Light speed in cm/s
C = LightSpeed(299792458. * 100)
#: Planck's constant in cm**2*g/s
H = Planck(6.6261e-27)

#: List of all atomic symbols
symbols = [
'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si',
 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co',
 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I',
 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U',
 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db',
 'Sg', 'Bh', 'Hs', 'Mt'
]

#: List of all atomic symbols in UPPERCASE
SYMBOLS = [x.upper() for x in symbols]




#!/usr/bin/env python

"""
Organizes simulation directory (creates folders, moves files, creates 'index.html')

  - moves 'root/report-*'       to 'root/reports'
  - moves 'root/C*'             to 'root/raw'
  - moves 'root/raw/simgroup*'  to 'root/'
  - moves 'root/raw/report-*'   to 'root/reports'
  - moves 'root/raw/group*.splist'   to 'root'
  - [re]creates 'root/reports/index.html'

This script can be run from one of these directories:
  - 'root' -- a directory containing at least one of these directories: 'reports', 'raw'
  - 'root/raw'
  - 'root/reports'

The script will use some rules to try to figure out where it is running from
"""

import argparse
import os
import glob
import sys
import f311.aosss as ao
import logging
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


class MoveTask(object):
    def __init__(self, source, dest):
        self.source = source
        self.dest = dest

    def __str__(self):
        cwd = os.getcwd()
        return "Move '{}' to '{}'".format(os.path.relpath(self.source, cwd),
                                          os.path.relpath(self.dest, cwd))

    def run(self):
        cwd = os.getcwd()
        dest_dir, _ = os.path.split(self.dest)
        if not os.path.exists(dest_dir):
            a99.get_python_logger().info("Creating directory '%s'..." % dest_dir)
            os.makedirs(dest_dir)
        a99.get_python_logger().info("Moving '{}' to '{}'".format(os.path.relpath(self.source, cwd),
                                                               os.path.relpath(self.dest, cwd)))
        os.rename(self.source, self.dest)


if __name__ == "__main__":
    lggr = a99.get_python_logger()

    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )

    args = parser.parse_args()

    # Tries to figure out where the root directory is
    rules = (
    ("reports", "."),
    ("raw", "."),
    ("../reports", ".."),
    ("../raw", ".."),
    )
    root = None
    for _dir, _root in rules:
        if os.path.isdir(_dir):
            root = _root
            lggr.info("Root directory is found to be '{}' because found directory '{}'".format(_root, _dir))
            break
    if root is None:
        if len(glob.glob("C*.fits")) > 0:
            lggr.info("Root directory is found to be '.' because found files 'C*.fits'")
            root = "."
        else:
            lggr.critical("Cannot determine simulations root directory, sorry")
            sys.exit()

    move_specs = (
    ('./report-*', './reports'),
    ('./C*', './raw'),
    ('./raw/simgroup-*', './'),
    ('./raw/report-*', './reports'),
    ('./raw/group*.splist', './'),
    )

    # Assembles a list of files to be moved
    to_move = []
    flag_reports = False
    for a, b in move_specs:
        files = glob.glob(os.path.join(root, a))
        for file in files:
            flag_reports = flag_reports or 'report-' in file
            full = os.path.abspath(file)
            if os.path.isfile(full):
                to_move.append(MoveTask(full,
                                        os.path.abspath(os.path.join(root, b, os.path.basename(file)))))

    # Determines whether there are reports to create index
    index_html = os.path.relpath(os.path.join(root, "reports", "index.html"))
    flag_reports = flag_reports or len(glob.glob(os.path.join(root, "reports", "report-*.html"))) > 0

    for task in to_move:
        lggr.info(task)

    lggr.info('Tasks Summary:')
    lggr.info(('  - Move {} object{}'.format(len(to_move), "s" if len(to_move) != 1 else "")))
    if flag_reports:
        lggr.info("  - Create '{}'".format(index_html))
    else:
        lggr.info("  - Will not create '{}' (no reports found)".format(index_html))

    if len(to_move) == 0 and not flag_reports:
        lggr.info('Nothing to do')
        sys.exit()

    flag_continue = False
    msg = "Continue (Y/n)? "
    while True:
        yn = input(msg)
        if yn == "":
            yn = "Y"
        if yn.upper() == "Y":
            flag_continue = True
            break
        elif yn.upper() == "N":
            break
        msg = "Please answer 'y' or 'n': "


    if not flag_continue:
        sys.exit()

    for task in to_move:
        task.run()

    if flag_reports:
        lggr.info("Creating '{0}'...".format(index_html))
        ao.create_index(os.path.join(root, "reports"))









__all__ = ["FilePar"]


import a99
from .. import DataFile
from collections import OrderedDict


@a99.froze_it
class FilePar(DataFile):
    """WebSim-COMPASS ".par" (parameters) file"""

    description = "Session parameters"
    default_filename = None
    attrs = ["params"]

    def __init__(self):
        DataFile.__init__(self)
        self.params = OrderedDict()

    def __getitem__(self, key):
        return self.params[key]

    def _do_load(self, filename):
        data = []  # keyword, value pairs create dict at the end
        # Opening with encoding other than 'utf-8' to avoid the following error:
        #
        #     UnicodeDecodeError: 'utf-8' codec can't decode byte 0xb0 in position 1782: invalid start byte
        with open(filename, "r", encoding='windows-1252') as f:
            i = 0
            try:
                for line in f:
                    s = line.strip()

                    if i == 0:
                        # Magic check: is it really a FilePar?
                        if not s.startswith("# parameter profile"):
                            raise RuntimeError("File '{}' is not a '.par' file".format(filename))

                    if s.startswith("#"):
                        pass
                    elif len(s) == 0:
                        pass
                    else:
                        KEYWORD_LENGTH = 26
                        keyword = s[:KEYWORD_LENGTH].strip()
                        value = s[KEYWORD_LENGTH:]
                        data.append((keyword, value))

                    i += 1
            except Exception as e:
                raise RuntimeError("In line {} of file '{}'".format(i+1, filename)) from e

        self.params = OrderedDict(data)

    def _do_save_as(self, filename):
        raise RuntimeError("Not applicable")




from .a_WFileAbonds import *
from .a_WFileAbXFwhm import *
from .a_WFileMain import *
from .a_WOptionsEditor import *
from .a_XAtomLinesEditor import *
from .a_XFileAbonds import *
from .a_XFileAtoms import *
from .a_XFileAtomsHistogram import *
from .a_XFileMain import *
from .a_XFileMolecules import *
from .a_XMainAbonds import *
from .a_XMolLinesEditor import *
from .a_WContinuum import *
from .a_WContinua import *
from .a_XFileOptions import *


from ... import filetypes as ft
from .basic import Vis
import matplotlib.pyplot as plt
import a99
import numpy as np

__all__ = ["VisGalfitFig", "draw_galfit_tiles"]


# TODO setup image size

class VisGalfitFig(Vis):
    """Draw tiles: (INPUT, MODEL, RESIDUAL) x (bands)"""

    input_classes = (ft.FileGalfit,)
    action = __doc__

    def _do_use(self, m):
        draw_galfit_tiles(m)
        plt.show()


def draw_galfit_tiles(filegalfit, image_width=1000):
    """Draws a new matplotilb figure and returns it

    The figure will contain  = 3*(number of bands) subplots

    Args:
        filegalfit: FileGalfit instance
        image_width=1000: image width in pixels

    Returns:
        matplotlib figure
    """
    # TODO fine-tune axis positions (not with tight_layout())

    num_rows = len(filegalfit.kind_names)
    num_cols = len(filegalfit.band_names)
    fig, axarr = plt.subplots(num_rows+1, num_cols, squeeze=False)
    for i, kind_name in enumerate(filegalfit.kind_names):
        for j, band_name in enumerate(filegalfit.band_names):
            ax = axarr[i, j]
            hdu = filegalfit.get_frame(kind_name, band_name)

            remove_ticks(ax)

            im = hdu.data
            # Clips negative values
            im[im < 0] =  0

            # Transforms color values to something that will enhance small values
            image_data = np.power(im, 0.2)
            hei, wid = image_data.shape

            ax.imshow(image_data, cmap='gray')
            ax.set_ylim([0, hei - 1])
            ax.set_xlim([0, wid - 1])

            if i == 0:
                # First row, will indicate band on top
                ax.text(wid/2, hei+4, filegalfit.band_names[j],
                        horizontalalignment='center',
                        verticalalignment='bottom')
                # plt.title(filegalfit.band_names[i])

            if j == num_cols-1:
                # Last column, will write kind_name as rotated text at the right of the tile
                ax.text(wid+4, hei/2, filegalfit.kind_names[i],
                        horizontalalignment='left',
                        verticalalignment='center',
                        rotation=90)

    if True:
        for j, band_name in enumerate(filegalfit.band_names):
            ax = axarr[3, j]
            hdu = filegalfit.get_frame("MODEL", band_name)



            for x in ["top", "left", "right", "bottom"]:
                ax.spines[x].set_visible(False)
            # ax.set_box("off")
            remove_ticks(ax)

            im = hdu.data
            hei, wid = hdu.data.shape

            ax.set_ylim([0, hei - 1])
            ax.set_xlim([0, wid - 1])

            CHI2NU = "\n\nCHI2NU={:g}".format(hdu.header["CHI2NU"]) if j == 0 else ""
            RE = hdu.header["2_RE_{}".format(band_name.upper())]
            N = hdu.header["2_N_{}".format(band_name.upper())]
            RS = hdu.header["3_RS_{}".format(band_name.upper())]
            ax.text(2, hei - 1, "RE={}\nN={}\nRS={}{}".format(RE, N, RS, CHI2NU),
                    horizontalalignment='left',
                    verticalalignment='top', fontsize=7, color='red',
                    bbox={'facecolor': 'white', 'alpha': 0.8, 'pad': 2})

    a99.set_figure_size(fig, image_width, image_width/num_cols*4)
    plt.tight_layout()


def remove_ticks(ax):
    # http://stackoverflow.com/questions/12998430/remove-xticks-in-a-matplot-lib-plot
    ax.xaxis.set_major_locator(plt.NullLocator())
    ax.yaxis.set_major_locator(plt.NullLocator())

#
# class VisGalfitFig(Vis):
#     """Draw figure of 3 x 5 tiles"""
#
#     input_classes = (ft.FileGalfit,)
#     action = __doc__
#
#     def _do_use(self, m):
#         draw_15_tiles(m.hdulist)
#         plt.show()
#
#
# def draw_15_tiles(hdulist, image_width=1000):
#     """Draws a new matplotilb figure and returns it
#
#     The figure will contain 15 = 3*5 subplots
#
#     Args:
#         hdulist: object obtained using astropy.io.fits.open. This represents
#                  a FITS file containing 20 frames.
#                  Frames from 2nd to 16th will be used
#         image_width=1000: image width in pixels
#
#     Returns:!
#         matplotlib figure
#     """
#
#     band_names = "ugriz"
#     row_names = ["Image", "Model", "Residual"]
#
#     # TODO fine-tune axis positions (not with tight_layout())
#
#
#     fig = plt.figure()
#     for i, hdu in enumerate(hdulist[1:16]):
#         plt.subplot(3, 5, i + 1)  # 1 + i // 5, 1 + i % 5)
#
#         # http://stackoverflow.com/questions/12998430/remove-xticks-in-a-matplot-lib-plot
#         plt.gca().xaxis.set_major_locator(plt.NullLocator())
#         plt.gca().yaxis.set_major_locator(plt.NullLocator())
#
#         im = hdu.data
#         # Clips negative values
#         im[im < 0] =  0
#         # Transforms color values to something that will enhance small values
#         image_data = np.power(im, 0.2)
#         hei, wid = image_data.shape
#
#         plt.imshow(image_data, cmap='gray')
#         plt.ylim([0, hei - 1])
#         plt.xlim([0, wid - 1])
#
#         if i < 5:
#             # First row, will indicate band on top
#             plt.gca().text(wid/2, hei+4, band_names[i],
#                     horizontalalignment='center',
#                     verticalalignment='bottom')
#             # plt.title(band_names[i])
#         if i % 5 == 4:
#             # Last column, will
#             # plt.gca().yaxis.set_label_position("right")
#             # plt.ylabel(row_names[int(i/5)])
#             plt.gca().text(wid+4, hei/2, row_names[int(i/5)],
#                     horizontalalignment='left',
#                     verticalalignment='center',
#                     rotation=90)
#
#
#
#     a99.set_figure_size(fig, image_width, image_width/5*3)
#     plt.tight_layout()



#!/usr/bin/env python
"""
Retrieves and prints a table of molecular constants from the NIST Chemistry Web Book.

To do so, it uses web scraping to navigate through several pages and parse the desired information
from the book web pages.

It does not provide a way to list the molecules yet, but will give an error if the molecule is not
found in the NIST web book.

Example:

    print-nist.py OH

**Disclaimer** This script may stop working if the NIST people update the Chemistry Web Book.

"""


import tabulate
import sys
import a99
import logging
import argparse
import f311.convmol as cm


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('formula', type=str, help='NIST formula', nargs=1)
    args = parser.parse_args()

    data, header, title = cm.get_nist_webbook_constants(args.formula)
    print("\n*** {} ***\n".format(title))
    print(tabulate.tabulate(data, header))





from .filemod import *
from .filedissoc import *
from .fileabonds import *
from .fileabsoru2 import *
from .fileabxfwhm import *
from .fileatoms import *
from .filehmap import *
from .filemain import *
from .filemolecules import *
from .fileoptions import *
from .filepartit import *
from .filetoh import *
from .filevald3 import *
from .filemoldb import *
from .filehitrandb import *
from .filekuruczmol import *
from .filepleztio import *
from .filefcf import *
from .filemolconsts import *
from .fileconfconvmol import *

# from . import filemod
# from . import filedissoc
# from . import fileabonds
# from . import fileabsoru2
# from . import fileabxfwhm
# from . import fileatoms
# from . import filehmap
# from . import filemain
# from . import filemolecules
# from . import fileoptions
# from . import filepartit
# from . import filetoh
# from . import filevald3
# from . import filekuruczmol



#!/usr/bin/env python

"""
Editor for molecules SQLite database

This application can edit files of class FileMolDB.
"""

import sys
import argparse
import a99
import logging
import os
import f311.filetypes as ft
import f311.explorer as ex


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = ft.FileMolDB.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='Molecules database file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None

    m = None
    if args.fn is not None:
        m = ft.FileMolDB()
        m.load(args.fn)
    app = a99.get_QApplication([])
    form = ex.XFileMolDB(None, [m,])
    form.showMaximized()
    sys.exit(app.exec_())










__ll__ = ["XFileSparseCube"]

from .a_WFileSparseCube import *
import a99
import f311.filetypes as ft
from ..a_XFileMainWindow import *


class XFileSparseCube(XFileMainWindow):

    def _add_stuff(self):
        self.setWindowTitle(a99.get_window_title("Data Cube Editor"))

        ce = self.ce = WFileSparseCube(self)

        self.pages.append(MyPage(text_tab="FileSparseCube editor",
         cls_save=ft.FileSparseCube, clss_load=(ft.FileSparseCube, ft.FileFullCube), wild="*.fits", editor=ce))




"""
Low-level maths def's
"""


__all__ = ["bc_rubber", "rubberband", "_rubber_pieces", "poly_baseline"]


import numpy as np


def bc_rubber(vx):
    """
    Convex Polygonal Line baseline correction

    Args:
        vx: vector

    Returns: vx-rubberband(vx)
    """

    return vx-rubberband(vx)


def rubberband(vx):
    """
    Convex polygonal line (aka rubberband)

    Args:
        vx: 1-D numpy array

    Returns: the rubberband: a 1-D numpy array with same shape as vx

    This function stretches a polygonal line from below vx. The vertices of this multi-segment line
    will touch troughs of vx without crossing vx

    This was inspired on -- but is not equivalent to -- OPUS SB_Rubberband baseline correction [1].
    However, this one is parameterless, whereas OPUS RBBC asks for a number of points.

    References:
        [1] Bruker Optik GmbH, OPUS 5 Reference Manual. Ettlingen: Bruker, 2004.
    """
    pieces = [[vx[0]]]
    _rubber_pieces(vx, pieces)
    rubberband = np.concatenate(pieces)

    return rubberband


def _rubber_pieces(x, pieces):
    """Recursive function that add straight lines to list. Together, these lines form the rubberband."""
    nf = len(x)
    l = np.linspace(x[0], x[-1], nf)
    xflat = x - l
    idx = np.argmin(xflat)
    val = xflat[idx]
    if val < 0:
        _rubber_pieces(x[0:idx + 1], pieces)
        _rubber_pieces(x[idx:], pieces)
    else:
        pieces.append(l[1:])


def poly_baseline(flux, order, epsilon=None, maxit=None):
    """"
    Polynomial baseline

    Args:
      vx: np 1D array ("flux")
      order: polynomial order
      epsilon: tolerance to stop iterations. If zero or no value given, will default to sqrt(1/30*num_points)
      maxit: if informed, will restrict the maximum number of iterations to this value

    Returns: the baseline vector

    Reference: Beier BD, Berger AJ. Method for automated background subtraction from Raman spectra containing known
    contaminants. The Analyst. 2009; 134(6):1198-202. Available at: http://www.ncbi.nlm.nih.gov/pubmed/19475148.

    **Converted from MATLAB (IRootLab)**
    """

    nf = len(flux)
    if epsilon is None:
        # epsilon will be compared to a vector norm. If we assume the error at all elements to have the same importance, it
        # the norm of the error will be something like sqrt(nf*error_i)
        # For the tolerance to be 1 when nf = 1500 (empirically found to work) the formula below is set.
        epsilon = np.sqrt(1/90*nf);

    x = np.arange(1, nf+1) # x-values to be powered as columns of a design matrix
    M = np.zeros((nf, order+1))
    for i in range(0, order+1):
        M[:, i] = x**i
    MM = np.dot(M.T, M)
    flag_first = True
    it = 0
    while True:
        # (29/03/2011) Least-Squares solution is faster than polyfit()

        # Original formula in MATLAB: yp = (M*(MM\(M'*vx')))'
        yp = np.dot(M, (np.linalg.solve(MM, np.dot(M.T, flux)))).T

        flux = np.min(np.vstack((flux, yp)), 0)

        if not flag_first:
            if np.linalg.norm(flux-y_previous, 2) < epsilon:
                break
        else:
            flag_first = False

        y_previous = flux

        it += 1
        if it >= maxit:
            break

    return yp





import f311.physics as ph
import f311.filetypes as ft
#from .. import convmol as cm
import a99
from .convlog import *
from collections import OrderedDict
import sys
import numpy as np
from .conv import *


__all__ = ["ConvPlez"]


# TODO This is incomplete: have to sort other things first. The main issue is whether we are going to extarct more than one system at once from one of these multiple-system files

class ConvPlez(Conv):
    """Converts Plez molecular lines data to PFANT "sets of lines" """

    def __init__(self, flag_hlf=False, flag_normhlf=False, flag_fcf=False, flag_quiet=False,
                 fcfs=None, *args, **kwargs):
        Conv.__init__(self, *args, **kwargs)
        self.flag_hlf = flag_hlf
        self.flag_normhlf = flag_normhlf
        self.flag_fcf = flag_fcf
        self.flag_quiet = flag_quiet
        self.fcfs = fcfs


    def _make_sols(self, lines):
        from f311 import convmol as cm

        def append_error(msg):
            log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

        # C     BAND (v',v'')=(VL,V2L)
        # C     WN: vacuum wavenumber   WL : air wavelength
        # C     J2L: lower state angular momentum number
        # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
        # C     ramo: branch, for P: ID=1, for R: ID=2
        # C     name: file name/ coisovLv2L
        # C     HL: Honl-London factor
        # C     FR: oscillator strength

        if not isinstance(self.lines, ft.FilePlezTiO):
            raise TypeError("Invalid type for argument 'lines': {}".format(type(lines).__name__))

        transition_dict = filemoldb.get_transition_dict()
        linedata = lines.get_numpy_array()

        trcols = linedata[["vup", "vlow", "state_from", "state_to"]]
        trset = np.unique(trcols)
        # trset = trcols.drop_duplicates()
        # trset["id_state"] = 0
        #
        #
        # for _, tr in trset.iterrows():
        #     state_from = tr["state_from"].decode("ascii")
        #     state_to = tr["state_to"].decode("ascii")
        #     try:
        #         molconsts = transition_dict[(molconsts["formula"], state_from, state_to)]
        #         tr["id_state"] = molconsts["id_state"]
        #     except KeyError as e:
        #         msg = "Will have to skip transition: '{}'".format(a99.str_exc(e))
        #         log.errors.append(msg)
        #         if not flag_quiet:
        #             a99.get_python_logger().exception(msg)
        #         continue



        sols = []

        S = molconsts.get_S2l()
        DELTAK = molconsts.get_deltak()
        fe = molconsts["fe"]

        # TODO of course this hard-wire needs change; now just a text for OH A2Sigma-X2Pi
        LAML = 0  # Sigma
        LAM2L = 1 # Pi

        if self.flag_hlf:
            formulas = ph.doublet.get_honllondon_formulas(LAML, LAM2L)
        log = MolConversionLog(len(lines))

        for tr in trset:
            state_from = tr["state_from"].decode("ascii")
            state_to = tr["state_to"].decode("ascii")
            try:
                molconsts = transition_dict[(molconsts["formula"], state_from, state_to)]
            except KeyError as e:
                msg = "Will have to skip transition: '{}'".format(a99.str_exc(e))
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)
                continue

            qgbd = self.calculate_qgbd(tr["vlow"])
            qqv = qgbd["qv"]
            ggv = qgbd["gv"]
            bbv = qgbd["bv"]
            ddv = qgbd["dv"]
            sol = ft.SetOfLines(tr["vup"], tr["vlow"], qqv, ggv, bbv, ddv, 1., state_from, state_to)
            sols.append(sol)

            mask = trcols == tr  # Boolean mask for linedata
            for i, line in enumerate(linedata[mask]):
                try:
                    wl = line["lambda_"]
                    J2l = line["Jlow"]
                    branch = line["branch"].decode("ascii")
                    gf = line["gf"]

                    if flag_normhlf:
                        k = 2./ ((2*S+1) * (2*J2l+1) * (2-DELTAK))
                    else:
                        k = 1.

                    if flag_hlf:
                        raise NotImplementedError("Hönl-London factors not implemented for Plez molecular lines file conversion")

                        # hlf = formulas[branch](line.J2l)
                        # gf_pfant = hlf*k

                    else:
                        gf_pfant = k*gf

                    if flag_fcf:
                        raise RuntimeError("Franck-Condon factors not implemented for Plez molecular lines file conversion")

                        # fcf = cm.get_fcf_oh(line.vl, line.v2l)
                        # gf_pfant *= fcf

                    sol.append_line(wl, gf_pfant, J2l, branch)


                except Exception as e:
                    msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), a99.str_exc(e))
                    log.errors.append(msg)
                    if not flag_quiet:
                        a99.get_python_logger().exception(msg)
                    continue


        return (sols, log)



__all__ = ["FileFullCube"]


from . import FullCube
from astropy.io import fits
import a99
# import f311.explorer as ex
from .. import DataFile


@a99.froze_it
class FileFullCube(DataFile):
    """
    FITS Data Cube ("full" opposed to "sparse")

    **Note** normally, the DataFile classes load operation reads all contents
    and closes the file, however this class keeps the file **open** as an
    astropy.io.fits.HDUList object in self.hdulist, because apparently many
    HDUList methods need the file open to work, even after calling
    HDUList.readall()
    """
    attrs = ['wcube']
    description = "WebSim Compass Data Cube (FITS file)"
    default_filename = "default.fullcube"
    flag_txt = False

    def __init__(self):
        DataFile.__init__(self)
        self.wcube = FullCube()
        self.hdulist = None

    def _do_load(self, filename):
        self.hdulist = fits.open(filename)
        self.wcube = FullCube(self.hdulist[0])
        self.filename = filename

    def _do_save_as(self, filename):
        if not self.wcube.flag_wavelengthed:
            raise RuntimeError("Cannot save before at least one pixel has been \"painted\"""")
        a99.overwrite_fits(self.wcube.hdu, filename)



__all__ = ["WFileSpectrumList"]

import collections
import copy
import matplotlib.pyplot as plt
from pylab import MaxNLocator
import numbers
import numpy as np
import os
import os.path
from itertools import product, combinations, cycle
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_WSpectrumList import *
import a99
import f311.filetypes as ft


class WFileSpectrumList(a99.WEditor):
    """
    FileSpectrumList editor widget.

    Args:
      parent=None
    """

    @property
    def menu_actions(self):
        return self.wsptable.menu_actions

    def __init__(self, parent):
        a99.WEditor.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        # Whether there is sth in yellow background in the Headers tab
        self.flag_header_changed = False
        self.obj_square = None

        # # Central layout
        lantanide = self.layout_editor
        a99.set_margin(lantanide, 0)
        self.setLayout(lantanide)

        # ## Horizontal splitter occupying main area: (options area) | (plot area)
        sp2 = self.splitter2 = QSplitter(Qt.Horizontal)
        lantanide.addWidget(sp2)

        # ## Widget left of horizontal splitter, containing (File Line) / (Options area)
        wfilett0 = keep_ref(QWidget())
        lwfilett0 = QVBoxLayout(wfilett0)
        a99.set_margin(lwfilett0, 0)

        # ### Tabbed widget occupying left of horizontal splitter (OPTIONS TAB)
        tt0 = self.tabWidgetOptions = QTabWidget(self)
        lwfilett0.addWidget(tt0)
        tt0.setFont(a99.MONO_FONT)
        tt0.currentChanged.connect(self.current_tab_changed_options)

        # #### Tab: Vertical Splitter between "Place Spectrum" and "Existing Spectra"
        spp = QSplitter(Qt.Vertical)
        tt0.addTab(spp, "&Spectra")


        # ##### Spectrum Collection Editor area
        wex = QWidget()
        lwex = QVBoxLayout(wex)
        a99.set_margin(lwex, 3)
        # ###
        # lwex.addWidget(keep_ref(QLabel("<b>Existing spectra</b>")))
        ###
        w = self.wsptable = WSpectrumList(self.parent_form)
        w.changed.connect(self.on_spectra_edited)
        lwex.addWidget(w)

        # ##### Finally...
        # spp.addWidget(sa0)
        spp.addWidget(wex)

        # #### Headers tab
        sa1 = keep_ref(QScrollArea())
        tt0.addTab(sa1, "&Header")
        sa1.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        sa1.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Widget that will be handled by the scrollable area
        w = keep_ref(QWidget())
        sa1.setWidget(w)
        sa1.setWidgetResizable(True)
        lscrw = QVBoxLayout(w)
        a99.set_margin(lscrw, 3)
        ###
        lscrw.addWidget(keep_ref(QLabel("<b>Header properties</b>")))

        ###
        b = keep_ref(QPushButton("Collect field names"))
        b.clicked.connect(self.on_collect_fieldnames)
        lscrw.addWidget(b)

        # Form layout
        lg = keep_ref(QGridLayout())
        a99.set_margin(lg, 0)
        lg.setVerticalSpacing(4)
        lg.setHorizontalSpacing(5)
        lscrw.addLayout(lg)
        ###
        lscrw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        # field map: [(label widget, edit widget, field name, short description, long description, f_from_f, f_from_edit), ...]
        map = self._map1 = []

        ###
        x = keep_ref(QLabel())
        y = self.edit_fieldnames = QPlainTextEdit()
        y.textChanged.connect(self.on_header_edited)
        x.setBuddy(y)
        map.append((x, y, "&Field names", "'header' information for each spectrum", "", lambda: self._f.splist.fieldnames,
                   lambda: self.edit_fieldnames.toPlainText()))
        ###

        for i, (label, edit, name, short_descr, long_descr, f_from_f, f_from_edit) in enumerate(map):
            # label.setStyleSheet("QLabel {text-align: right}")
            assert isinstance(label, QLabel)
            label.setText(a99.enc_name_descr(name, short_descr))
            label.setAlignment(Qt.AlignRight)
            lg.addWidget(label, i, 0)
            lg.addWidget(edit, i, 1)
            label.setToolTip(long_descr)
            edit.setToolTip(long_descr)

        lgo = QHBoxLayout()
        a99.set_margin(lgo, 0)
        lscrw.addLayout(lgo)
        ###
        bgo = self.button_revert = QPushButton("Revert")
        lgo.addWidget(bgo)
        bgo.clicked.connect(self.header_revert)
        ###
        bgo = self.button_apply = QPushButton("Apply")
        lgo.addWidget(bgo)
        bgo.clicked.connect(self.header_apply)
        ###
        lgo.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ### Finally ...
        sp2.addWidget(wfilett0)
        # sp2.addWidget(tt1)


        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True
        self.add_log("Welcome from %s.__init__()" % (self.__class__.__name__))

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def _do_load(self, x):
        assert isinstance(x, (ft.FileSpectrum, ft.FileSpectrumList, ft.FileFullCube))

        # Converts from FileFullCube to FileSpectrumList format, if necessary
        x1 = None
        if isinstance(x, ft.FileFullCube):
            x1 = ft.FileSpectrumList()
            x1.splist.from_full_cube(x.wcube)
        elif isinstance(x, ft.FileSpectrum):
            x1 = ft.FileSpectrumList()
            x1.splist.add_spectrum(x.spectrum)
        if x1:
            x1.filename = a99.add_bits_to_path(x.filename, "imported-from-",
             os.path.splitext(ft.FileSpectrumList.default_filename)[1])
            x = x1

        self._f = x
        self.wsptable.set_collection(x.splist)
        self.__update_gui(True)
        self._flag_valid = True  # assuming that file does not come with errors
        self.setEnabled(True)

    def update_splist_headers(self, splist):
        """Updates headers of a SpectrumList objects using contents of the Headers tab"""
        emsg, flag_error = "", False
        ss = ""
        flag_emit = False
        try:
            ss = "fieldnames"
            ff = a99.eval_fieldnames(str(self.edit_fieldnames.toPlainText()))
            splist.fieldnames = ff
            self.__update_gui(True)
            flag_emit = True
        except Exception as E:
            flag_error = True
            if ss:
                emsg = "Field '%s': %s" % (ss, a99.str_exc(E))
            else:
                emsg = a99.str_exc(E)
            self.add_log_error(emsg)
        if flag_emit:
            self.__emit_if()
        return not flag_error


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_colors_setup_edited(self):
        if self.flag_process_changes:
            pass
            # self.flag_plot_colors_pending = True

    def on_header_edited(self):
        if self.flag_process_changes:
            sth = False
            sndr = self.sender()
            for _, edit, _, _, _, f_from_f, f_from_edit in self._map1:
                changed = f_from_f() != f_from_edit()
                sth = sth or changed
                if edit == sndr:
                    a99.style_widget_changed(self.sender(), changed)
            self.set_flag_header_changed(sth)

    def add_spectrum_clicked(self):
        flag_emit = False
        try:
            sp = self.choosesp.sp
            if not sp:
                raise RuntimeError("Spectrum not loaded")
            sp = copy.deepcopy(sp)
            self._f.splist.add_spectrum(sp)
            self.__update_gui()
            flag_emit = True
        except Exception as E:
            self.add_log_error(a99.str_exc(E), True)
            raise
        if flag_emit:
            self.changed.emit()

    def header_revert(self):
        self.__update_gui_header()

    def header_apply(self):
        if self.update_splist_headers(self._f.splist):
            self.__update_gui(True)

    def current_tab_changed_vis(self):
        pass
        # if self.flag_plot_colors_pending:
        #     self.plot_colors()

    def current_tab_changed_options(self):
        pass

    def on_collect_fieldnames(self):
        # TODO confirmation

        self.edit_fieldnames.setPlainText(str(self._f.splist.collect_fieldnames()))

    #        self.__update_gui(True)

    def on_spectra_edited(self, flag_changed_header):
        if flag_changed_header:
            self.__update_gui_header()
        self.__update_gui_vis()
        self.changed.emit()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __emit_if(self):
        if self.flag_process_changes:
            self.changed.emit()

    def __update_gui(self, flag_header=False):
        """Updates GUI to reflect what is in self._f"""
        self.flag_process_changes = False
        try:
            self.wsptable.update()
            if flag_header:
                self.__update_gui_header()
        finally:
            self.flag_process_changes = True

    def __update_gui_header(self):
        """Updates header controls only"""
        splist = self._f.splist
        self.edit_fieldnames.setPlainText(str(splist.fieldnames))
        self.set_flag_header_changed(False)

    def __update_gui_vis(self):
        pass

    def set_flag_header_changed(self, flag):
        self.button_apply.setEnabled(flag)
        self.button_revert.setEnabled(flag)
        self.flag_header_changed = flag
        if not flag:
            # If not changed, removes all eventual yellows
            for _, edit, _, _, _, _, _ in self._map1:
                a99.style_widget_changed(edit, False)

    def __update_f(self):
        self._flag_valid = self.update_splist_headers(self._f.splist)

    def __new_window(self, clone):
        """Opens new FileSparseCube in new window"""
        form1 = self.keep_ref(self.parent_form.__class__())
        form1.load(clone)
        form1.show()

    def __use_sblock(self, block):
        """Uses block and opens result in new window"""
        from f311 import explorer as ex

        # Does not touch the original self._f
        clone = copy.deepcopy(self._f)
        clone.filename = None
        slblock = ex.SLB_UseSpectrumBlock()
        for i, sp in enumerate(clone.splist.spectra):
            clone.splist.spectra[i] = block.use(sp)
        self.__new_window(clone)

    def __use_slblock(self, block):
        """Uses block and opens result in new window"""
        # Here not cloning current spectrum list, but trusting the block
        block.flag_copy_wavelength = True
        output = block.use(self._f.splist)
        f = self.__new_from_existing()
        f.splist = output
        self.__new_window(f)

    def __new_from_existing(self):
        """Creates new FileSpectrumList from existing one"""
        f = ft.FileSpectrumList()
        return f




#!/usr/bin/env python

"""
Editor for molecular constants file

This application can edit files of class FileMolConsts.
"""

import sys
import argparse
import a99
import logging
import os
import f311.filetypes as ft
import f311.explorer as ex


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = ft.FileMolConsts.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='Molecular constants file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None


    # Currently making copy of FileMolDB database and deleting afterwards
    moldb = ft.FileMolDB()
    moldb.init_default()
    try:
        fobj = None
        if args.fn is not None:
            fobj = ft.FileMolConsts()
            fobj.load(args.fn)
        app = a99.get_QApplication([])
        form = ex.XFileMolConsts(None)
        form.set_moldb(moldb)
        form.load(fobj)
        form.show()
    finally:
        try:
            os.unlink(moldb.filename)
        except FileNotFoundError:
            pass  # nevermind

    sys.exit(app.exec_())










__all__ = [
"create_spectrum_lists", "load_bulk", "BulkItem", "load_eso_sky"]


import os.path
from collections import OrderedDict
import glob
import os
import re
import numpy as np
import a99
import f311.filetypes as ft
from astropy.io import fits
import astropy.units as u
import f311.explorer as ex
# from .. import aosss as ao


FILE_MAP = OrderedDict((
 ("cube_hr", ft.FileFullCube),
 ("cube_seeing", ft.FileFits),
 ("ifu_noseeing", ft.FileFullCube),
 ("mask_fiber_in_aperture", ft.FileFits),  # TODO, handle this file because it is nice
 ("reduced", ft.FileFullCube),
 ("reduced_snr", ft.FileFullCube),
 ("sky", ft.FileSpectrumFits),
 ("skysub", ft.FileSpectrumFits),
 ("spintg", ft.FileSpectrumFits),
 ("spintg_noseeing", "messed"),  # particular case
 ("therm", "messed"),            # particular case
))
#
# FILE_KEYWORDS = FILE_MAP.keys()


class BulkItem(object):
    """This is each item returned by load_bulk"""
    def __init__(self, keyword=None, fileobj=None, filename=None, flag_exists=None,
                 flag_supported=None, error=None):
        # fileobj may be a DataFile or a string (the latter if the .par file)
        self.keyword = keyword
        self.fileobj = fileobj
        self.filename = filename
        self.flag_exists = flag_exists
        self.flag_supported = flag_supported
        self.error = error


def load_bulk(simid, dir_='.'):
    """
    Loads all files given the simulation id. Returns an list of BulkItem

    The keys are the same as in FILE_KEYWORDS, in the same order as well
    """

    ret = []
    sp_ref = None  # reference spectrum for x-axis of messed files
    for keyword, class_ in list(FILE_MAP.items()):
        fn = os.path.join(dir_, "%s_%s.fits" % (simid, keyword))

        flag_exists = os.path.isfile(fn)
        flag_supported = class_ is not None
        error = None
        fileobj = None
        if flag_exists:
            if class_ == "messed":
                # Note that sp_ref may be None here if Cxxxx_spintg.fits fails to load
                # In this case, sp will have a default x-axis
                sp = ft.load_spectrum_fits_messed_x(fn, sp_ref)
                if sp:
                    fileobj = ft.FileSpectrum()
                    fileobj.spectrum = sp
            elif flag_supported:
                fileobj = class_()
                try:
                    fileobj.load(fn)
                except Exception as E:
                    a99.get_python_logger().exception("Error loading file '%s'" % fn)
                    error = str(E)

                if keyword == "spintg":
                    # Saves reference spectrum to use its x-axis to complete
                    # the "messed" files
                    sp_ref = fileobj.spectrum

            ret.append(BulkItem(keyword, fileobj, fn, flag_exists, flag_supported,
                            error))
    return ret


def create_spectrum_lists(dir_, pipeline_stage="spintg"):
    """
    Create several .splist files, grouping spectra by their wavelength vector

    Args:
        dir_: input & output directory
        pipeline_stage="spintg": input "stage", i.e., which stage of the pipeline will be loaded.
            Possible values:

            - "spintg": integrated spectrum (final stage)
            - "ifu_noseeing": take first spectrum of 7x1 data cube TODO explain
    """

    if pipeline_stage in ("spintg", "sky", "skysub"):

        def simid_to_spectrum(simid):
            fn = os.path.join(dir_, simid + "_{}.fits".format(pipeline_stage))
            fsp = ft.FileSpectrumFits()
            fsp.load(fn)
            return fsp.spectrum

    elif pipeline_stage == "ifu_noseeing":

        def simid_to_spectrum(simid):
            fn = os.path.join(dir_, simid + "_ifu_noseeing.fits")
            fsp = ft.FileFullCube()
            fsp.load(fn)
            return fsp.spectrum

    else:
        raise ValueError("Invalid or unsupported pipeline stage: '{}'".format(pipeline_stage))

    fnfn = glob.glob(os.path.join(dir_, "C*.par"))
    fnfn.sort()
    # # Loads everything that's needed from disk
    spectra = []  # [(FilePar0, Spectrum0), ...]
    for fn in fnfn:
        try:
            gg = re.search('C(\d+)', fn)
            if gg is None:
                raise RuntimeError(
                    "'.par' file name '%s' does not have pattern 'Cnnnnnn'" % fn)

            fp = ft.FilePar()
            fp.load(fn)

            sp = simid_to_spectrum(gg.group())

            spectra.append((fp, sp))
        except:
            a99.get_python_logger().exception(
                "Failed to add spectrum corresponding to file '%s'" % fn)

    # Groups files by their wavelength axis
    groups = []  # list of lists: [[(FilePar0, Spectrum0), ...], ...]
    for pair in spectra:
        if len(groups) == 0:
            groups.append([pair])
        else:
            flag_match = False
            for group in groups:
                wl0 = group[0][1].wavelength
                wl1 = pair[1].wavelength
                if len(wl0) == len(wl1) and np.all(wl0 == wl1):
                    flag_match = True
                    group.append(pair)
                    break
            if not flag_match:
                groups.append([pair])

    # Now creates the spectrum list files
    for h, group in enumerate(groups):

        # # Finds differences in simulation specifications
        # Finds names of parameters that differ at least in one file using set operations
        for i, (fp, _) in enumerate(group):
            s = set(fp.params.items())
            if i == 0:
                s_union = s
                s_overlap = s
            else:
                s_union = s_union | s
                s_overlap = s_overlap & s
        keys = list(dict(s_union - s_overlap).keys())
        key_dict = a99.make_fits_keys_dict(keys)

        # a99.get_python_logger().info("FITS headers to feature in all spectra:")
        # a99.get_python_logger().info(str(key_dict))


        fspl = ft.FileSpectrumList()
        nmin, nmax = 9999999, 0
        for fp, sp in group:
            try:
                gg = re.search('C(\d+)', fp.filename)
                if gg is None:
                    raise RuntimeError(
                        "'.par' file name '%s' does not have pattern 'Cnnnnnn'" % fp.filename)
                n = int(gg.groups()[0])
                nmin = min(n, nmin)
                nmax = max(n, nmax)

                # determines what to put in "ORIGIN" header
                origin = sp.more_headers.get("ORIGIN")
                origin = origin if origin else os.path.basename(sp.filename)

                # gets rid of all FITS headers (TODO not sure if this is a good idea yet)
                sp.clear_more_headers()
                sp.more_headers["ORIGIN"] = origin

                # copies to spectrum header all key-value pairs that differ among .par files
                for k in keys:

                    # unwanted parameters
                    # - simu name is redundant with simu_id
                    # - ext_link has info such as "Resource id #4"
                    if k in ["ext_link", "simu_name"]:
                        # TODO not efficient to have this here, should remove from keys
                        continue

                    value = fp.params.get(k)
                    if value is not None:
                        # Content-sensitive conversion
                        if k == "obj_ftemplate":
                            # Template filename comes with full local
                            # (i.e. Websim server) path, which is unnecessary
                            # for our identification
                            value = os.path.basename(value)
                        elif k == "psf_file":
                            # PSF filename also comes with its full path,
                            # which if probably irrelevant
                            value = os.path.basename(value)
                        # else:
                        #       # .get(k)

                    sp.more_headers[key_dict[k]] = value

                fspl.splist.add_spectrum(sp)
            except:
                a99.get_python_logger().exception(
                    "Failed to add spectrum corresponding to file '%s'" % fp.filename)

        fn = os.path.join(dir_, "group-%s-%02d-C%06d-C%06d.splist" % (pipeline_stage, h, nmin, nmax))
        fspl.save_as(fn)
        a99.get_python_logger().info("Created file '%s'" % fn)


def load_eso_sky():
    """Loads ESO sky model from data directory

    Returns:
        tuple: ``(emission, transmission)`` (two `f311.filetypes.Spectrum` objects)
    """
    from f311 import aosss as ao

    # From comments in file:
    # lam:     vacuum wavelength in micron
    # flux:    sky emission radiance flux in ph/s/m2/micron/arcsec2
    # dflux1:  sky emission -1sigma flux uncertainty
    # dflux2:  sky emission +1sigma flux uncertainty
    # dtrans:  sky transmission
    # dtrans1: sky transmission -1sigma uncertainty
    # dtrans2: sky transmission +1sigma uncertainty

    path_ = a99.get_path("data", "eso-sky.fits", module=ao)

    hl = fits.open(path_)
    d = hl[1].data

    x, y0, y1 = d["lam"]*10000, d["flux"], d["trans"]

    sp0 = ft.Spectrum()
    sp0.x, sp0.y = x, y0
    sp0.yunit = u.Unit("ph/s/m2/angstrom/arcsec2")

    sp1 = ft.Spectrum()
    sp1.x, sp1.y = x, y1

    return sp0, sp1



import a99
from ..basic import Vis
import f311.filetypes as ft


__all__ = ["VisCube", "VisSpectrumList"]


class VisCube(Vis):
    """Opens the Data Cube Editor window."""
    input_classes = (ft.FileFullCube, ft.FileSparseCube)
    action = "Edit using Data Cube Editor"

    def _do_use(self, r):
        from f311 import explorer as ex
        form = a99.keep_ref(ex.XFileSparseCube(self.parent_form, r))
        form.show()


class VisSpectrumList(Vis):
    """Opens the Spectrum List Editor window."""
    input_classes = (ft.FileSpectrumList, ft.FileSpectrum)
    action = "Edit using Spectrum List Editor"

    def _do_use(self, r):
        from f311 import explorer as ex
        form = a99.keep_ref(ex.XFileSpectrumList(self.parent_form, r))
        form.show()



__all__ = ["XUseSpectrumBlock"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XHelpDialog import *
import a99


_CONFIG_OPERATIONS = "/gui/XUseSpectrumBlock/operations"


class XUseSpectrumBlock(XHelpDialog):
    """
    Edit Parameters to apply a SpectrumBlock to each spectrum in a Spectrum List

    Relevant attributes (set on close):
      self.block: None or SpectrumBlock instance
    """

    def __init__(self, *args):
        from f311 import explorer as ex

        XHelpDialog.__init__(self, *args)

        self.previous_operations = ex.get_config().get_item(_CONFIG_OPERATIONS, [])
        self.block = None

        self.setWindowTitle("Transform")

        self.labelHelpTopics.setText("&Available operations")

        self.help_data = a99.collect_doc(ex.sb, base_class=ex.SpectrumBlock)
        self.comboBox.addItems([x[0] for x in self.help_data])
        ###
        label = QLabel(a99.enc_name_descr("O&peration", "See help below"))
        label.setAlignment(Qt.AlignRight)
        cb = self.cb_operation = QComboBox()
        label.setBuddy(cb)
        cb.setEditable(True)
        cb.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        cb.addItems(self.previous_operations)
        self.grid.addWidget(label, 0, 0)
        self.grid.addWidget(cb, 0, 1)

        a99.place_center(self, 800, 600)


    def accept(self):
        from f311 import explorer as ex

        try:
            expr = str(self.cb_operation.currentText())
            block = eval(expr.strip(), {}, a99.module_to_dict(ex.blocks.sb))

            if not isinstance(block, ex.SpectrumBlock):
                raise RuntimeError("Expression does not evaluate to a valid Spectrum Block")

            self.block = block

            # Saves new operation on close (if evaluated successfully)
            if not expr in self.previous_operations:
                self.previous_operations.append(expr)
                self.previous_operations.sort()
                ex.get_config().set_item(_CONFIG_OPERATIONS, self.previous_operations)

            return QDialog.accept(self)
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True)
            return False
